// Object that creates a shadow DOM for and SVG on the parent container
// and handles keyboard input
function svgShadowDom(parentId) {
	this.id = '#' + parentId; // jQuery Id for the parent container
	this.$par = $(this.id);

	this.keys = {
		tab : 9,
		enter : 13,
		escape : 27,
		space : 32,
		pageup : 33,
		pagedown : 34,
		end : 35,
		home : 36,  
		left : 37,
		up : 38,
		right : 39,
		down : 40,
		e : 54,
		nw : 55,
		n : 56,
		ne : 57,
		w : 58,
		sw : 49,
		se : 51,
		s : 50,
		n1 : 97,
		n2: 98,
		n3: 99,
		n4: 100,
		n6: 102,
		n7: 103,
		n8: 104,
		n9: 105
	};
	
	// Modify the parent container to be navigable and make JAWS report changes
	this.$par.attr('tabindex', '0');
	this.$par.attr('aria-live', 'assertive');
	this.$par.attr('role', 'application');

	// bind keydown handler to parent
	var thisObj = this;
	this.$par.keydown(function(e) {
		return thisObj.handleKeyDown(e);
	});
}

// Removes an old shadow DOM and builds a new shadow DOM
svgShadowDom.prototype.init = function() {
	
	this.$par.remove('.svgShadowDom'); // Remove any existing shadow DOM	
	this.sdElemWithFocus = null;	   // Clear shadow DOM element with focus
	
	this.$svg = this.$par.find('svg'); // The SVG belonging to the parent container
	
	// A tree of sdItem's are built for the semantic (navigable) SVG elements
	// then the sdItems build the shadow DOM
	// The sdItem allow us to tweak organization - like using graphics-titles in computing the
	// name of the parent group and excluding graphics-title as a child (which is a pain to navigate to)
	
	// Elements with roles are semantic (and navigable), unless role is none 
	var $roleNone = this.$svg.find("[role='none']");
	var $withRoles = this.$svg.find('[role]').not($roleNone);

	// Elements with a first child of <title> or <desc> are semantic, unless role of none
	var $hasTitles = this.$svg.find(':first-child').filter('title').parent();
	var $hasDesc = this.$svg.find(':first-child').filter('desc').parent();

	// Elements with aria-(label, labeledby, describedby) are semantic, unless role of none
	var $labeledBy = this.$svg.find('[aria-labeledby]');
	var $describedBy = this.$svg.find('[aria-describedby]');
	var $labels = this.$svg.find('[aria-label]');

	// Everything semantic (navigable) in one list
	var $all = $withRoles.add($describedBy).add($labels).add($labeledBy).add($hasTitles).add($hasDesc).not($roleNone);

	// Groups with semantic meaning, groups used to build the tree 
	this.$groups = $all.filter('g');
	this.$groups = this.$groups.add(this.$svg.find('svg'));
	this.$navigableItems = $all;
	
	this.dmap = new dmap();  // Map for objects, key is DOM node, value is sdItem.
	
	// Add atomic items to dmap, atomic items will be leaves in the tree (no children)
	var $atomicGroups = this.mapAtomicItems($all);
	
	// Build a tree of sdItems - the tree root will build the shadow DOM
	this.buildSdTree($atomicGroups);
	var svg = this.$svg.get(0);  // SVG DOM element
	var svgSdItem = this.dmap.get(svg);
	// this.root is used in handleKeydown
	this.root = svgSdItem.build(this.$par.get(0), true, svg);
	
	// Cleanup
	this.dmap = null;
	this.$groups = null;
	this.$navigableItems = null;
	
} // end init()

// Creates an sdItem for each atomic item, puts them in this.dmap
svgShadowDom.prototype.mapAtomicItems = function($all) {

	// Find the atomic items
	// Find groups with atomic roles
	var $atomicGroups = this.$groups.filter("[role='symbol']");
	$atomicGroups = $atomicGroups.add(this.$groups.filter("[role='graphics-label']")).add(this.$groups.filter("[role='graphics-axistick']"));
	
	// Atomic items
	var $atomicItems = $all.not(this.$groups).not(this.$svg);
	var $atomics = $atomicItems.add($atomicGroups);
	
	// Create an sdItem for each atomic item and add to map
	var length = $atomics.length;
	for (var i = 0; i < length; i++) {
		var item = $atomics.get(i);
		var navItem = new sdItem(item, null);
		this.dmap.put(item, navItem);
	}

	return $atomicGroups;  // Return groups with atomic roles
}

// Builds a tree of sdItem
svgShadowDom.prototype.buildSdTree = function($atomicGroups) {
	var $used = $();  // jQuery object of 'used' DOM elements
	
	// jQuery puts DOM nodes in DOM order in jQuery objects
	// We will go last to first through the groups building the tree for the graphics
	// the backward order ensures all the children for a branch are in the dmap when the 
	// item for the group (branch) is built
	var $semanticGroups = this.$groups.not($atomicGroups);
	var $semanticGroups = $semanticGroups.add(this.$svg);
	var max = $semanticGroups.length - 1;
	for (var i = max; i > -1; i--) {
		$descendants = this.buildBranch($semanticGroups.get(i), $used);
		$used = $used.add($descendants);
	}
		
}

// Builds an sdItem for a group, the sdItem is a branch in the tree
svgShadowDom.prototype.buildBranch = function(g, $used) {
	var $g = $(g);
	// $descendants are the leaves and sub branches in this branch
	// $used are excluded as they have already been 'used' by a sub branch
	var $descendants = $g.find(this.$navigableItems).not($used); 
	
	// DESIGN DECISION
	// I've decided to use graphics-title in the parent's name calculation and not include it in navigation
	// this simplifies navigation and makes the parent group's name similar to reading the title
	// putting graphics-title in sdItem.gtitle and excluding from the branch members
	var $gTitle = $descendants.filter("[role='graphics-title']");
	var $members = $descendants.not($gTitle);
	// second param is array of sdItem that are the children of this branch
	var branch = new sdItem(g, this.dmap.getArray($members.get()));
	
	if ($gTitle.length > 0) {
		// Setting the graphics-title in gtitle for use in name calculation
		branch.gtitle = this.dmap.get($gTitle.get(0));
	}

	this.dmap.put(g, branch);  // Add this group to the map, the group's parent will need it
	return $descendants; 	   // Returns the DOM elements used in this group
}

// Sets the focus on the selected shadow DOM element
svgShadowDom.prototype.setFocusOn = function(selected) {
	this.sdElemWithFocus= selected;
	if (selected) {
		selected.focus();
	}	
//	this.debugMsg(selected);
}
//svgShadowDom.prototype.debugMsg = function(node) {
//	var debug = document.getElementById('debugTextMsg');
//	if (!node) {
//		debug.innerHTML = '---';
//		return;
//	}
//	
//	var name = node.tagName;
//	var tc = node.textContent;
//	if (tc) {
//		tc = (tc.length > 50) ? tc.substring(0, 50) : tc;
//	} else {
//		tc = ' ';
//	}
//	
//	debug.innerHTML = name + " " + tc;
//}

svgShadowDom.prototype.siblingMove = function(down) {
	var sibling = null;
	if (this.sdElemWithFocus.className == 'svgShadowDom') {
		// Can't use up or down arrow key to move from the shadow DOM element for the SVG element
		return;
	} else if (this.sdElemWithFocus.nodeName.toLowerCase() == 'li') {
		sibling = (down) ? this.sdElemWithFocus.nextElementSibling : this.sdElemWithFocus.previousElementSibling;
		
	} else if (this.sdElemWithFocus.nodeName.toLowerCase() == 'ul') {
		var parent = this.sdElemWithFocus.parentElement;
		sibling = (down) ? parent.nextElementSibling : parent.previousElementSibling;
	}
	
	if (sibling && sibling.nodeName.toLowerCase() == 'li') {
		var child = sibling.firstElementChild;
		if (child && child.nodeName.toLowerCase() == 'ul') {
			sibling = child;
		}				
	}		
	if (sibling) {
		this.setFocusOn(sibling);
	}
	
}

svgShadowDom.prototype.handleKeyDown = function(e) {

	var caseFound = true;

	switch (e.keyCode) {
	case this.keys.left:
	case this.keys.w: {
		if (this.sdElemWithFocus == null) {
			this.setFocusOn(this.root);
		} else {
			var parent = this.sdElemWithFocus.parentElement;
			if (this.sdElemWithFocus.childElementCount > 0) {
				this.sdElemWithFocus.setAttribute('aria-expanded', false);
			}			
			this.setFocusOn(parent);

		}
		break;
	}
	case this.keys.right:
	case this.keys.e: {
		// open child
		if (this.sdElemWithFocus == null) {
			this.setFocusOn(this.root);
		} else {
			var child = this.sdElemWithFocus.firstElementChild;
			if (child) {
				this.sdElemWithFocus.setAttribute('aria-expanded', true);
				this.setFocusOn(child);
			}			
		}
		
		break;
	}
	case this.keys.up:
	case this.keys.n: {
		if (this.sdElemWithFocus == null) {
			this.setFocusOn(this.root);
		} else {
			this.siblingMove(false);	
		}
	}
	break;
	case this.keys.down:
	case this.keys.s: {
		if (this.sdElemWithFocus == null) {
			this.setFocusOn(this.root);
		} else {
			this.siblingMove(true);	
		}

	}
	break;
	case this.keys.home: {
		var $sdRoot = $(this.root);
		$sdRoot.find("[aria-expanded='true']").attr('aria-expanded', 'false');
		this.setFocusOn(this.root);
		break;
	}
	default:
		caseFound = false;
	}

	if (caseFound) {
		e.stopPropagation();
	}
	return !caseFound;

} // end handleKeyDown

// Object to build the shadow DOM
function sdItem(i, childArray) {
	this.item = i;
	this.children = childArray;
	
	// Capture a title and desc, will use later in building the name (text) the screen reader will get
	var $g = $(i);
	var $title = $g.children('title');
	if ($title.length > 0) {
		this.title = new sdItem($title.get(0), null);
	}
	var $desc = $g.children('desc');
	if ($desc.length > 0) {
		this.desc = new sdItem($desc.get(0), null);
	}	
}

sdItem.prototype.build = function(parent, root, svg) {
	var hasChild = (this.children && this.children.length > 0);
	var elem;
	if (root) {
		elem = this.createElement('ul');
		elem.setAttribute('role', 'tree');
		elem.setAttribute('class', 'svgShadowDom');
		elem.setAttribute('aria-expanded', 'true');
		elem.setAttribute('tabindex', '0');
		parent.insertBefore(elem, svg);
		// TODO -- to move the shadow DOM off screen, uncomment out the next two lines
//		elem.style.position = 'absolute';
//		elem.style.left = '-3000px';
	} else {
		elem = this.createElement('li');
		elem.setAttribute('role', 'treeitem');
		elem.setAttribute('tabindex', '-1');
		parent.appendChild(elem);				
	}
	
	if(hasChild) {
		var childParent = elem;
		if (!root) {
			childParent = document.createElement('ul');
			childParent.setAttribute('aria-expanded', 'false');
			childParent.setAttribute('tabindex', '-1');
			childParent.setAttribute('role', 'group');
			childParent.setAttribute('aria-label', elem.getAttribute('aria-label'));
			elem.appendChild(childParent);
		}
		
		for(var i = 0; i < this.children.length; i++) {
			this.children[i].build(childParent, false);
		}
		
	}
	return elem;
}

sdItem.prototype.addValue = function(attr, text) {
	var value = this.item.getAttribute(attr);
	if (value) {
		var name = attr;
		if (attr.indexOf('aria-valuem') > -1) {
			name = attr.substring(10);
		} else if (attr.indexOf('aria-') > -1) {
			name = attr.substring(5);
		} else if (attr.indexOf('graphics-') > -1) {
			name = attr.substring(8);
		}
		text = text + ' ' + name + ': ' + value;
	}
	return text;
}


sdItem.prototype.nameFromItem = function() {
	
	// Use the aria-label
	if (this.item.hasAttribute('aria-label')) {
		return this.item.getAttribute('aria-label');
	} 
	// Use the title
	if (this.title) {
		return this.title.item.textContent;
	} 
	
	// Calculate the name
	var role = this.item.getAttribute('role');
	var isAxis = (role == 'graphics-axis');

	var text;
	if (isAxis || (this.item.children && this.item.children.length > 0)) {
		text = this.nameFromRole(this.item);
	} else {
		text = this.item.textContent;
	}
	
	return this.addAriaProperties(text);
}

sdItem.prototype.addAriaProperties = function(text) {
	if (this.gtitle) {
		text = text + " " + this.gtitle.nameFromItem();
	}
	text= this.addValue('aria-valuemin', text);
	text= this.addValue('aria-valuemax', text);
	return this.addValue('aria-categories', text);						
}

sdItem.prototype.createElement = function(tagName) {
	// Create a DOM element
	var element = document.createElement(tagName);
	var item = this.item;
	if (item.hasAttribute('class')) {
		// Save the item's class will swap it out when shadow DOM element has focus, keep original to restore
		element.setAttribute('origClass', item.getAttribute('class'));
	}
		
	if (item.hasAttribute('aria-labeledby')) {
//		element.setAttribute('aria-labeledby', item.getAttribute('aria-labeledby'));
		var name = this.getTextFromReference(item, 'aria-labeledby');		
		name = this.addDescription(name);
		element.textContent = name;
		element.setAttribute('aria-label', name);
	} else {	
		// If no aria-labeledby, get the name from the item, this may do calculations
		var name = this.nameFromItem();	
		name = (name) ? name : '';  // An SVG can end with no name if it doesn't have a title or role
		name = this.addDescription(name);
		element.textContent = name;
		element.setAttribute('aria-label', name);
	}
		
	// save the SVG DOM node for showing pseudo focus
	element['svgNode'] =  this.item;
	element.onfocus = function() {
		var svgNode = this['svgNode'];
		svgNode.setAttribute('class', 'activeElem');
	};
	
	element.onblur = function() {
		var svgNode = this['svgNode'];
		if (this.hasAttribute('origClass')) {
			var origClass = this.getAttribute('origClass');
			svgNode.setAttribute('class', origClass);
		} else {
			svgNode.removeAttribute('class');
		}
		
	};
	
	return element;
}

sdItem.prototype.getTextFromReference = function(item, attr) {
	if (item.hasAttribute(attr)) {
		var id = '#' + item.getAttribute(attr);
		var textItem = $(id).get(0);
		// Chrome doesn't like textContent on text elements
		return  (textItem.nodeName.toLowerCase() == 'text') ? textItem.innerHTML : textItem.textContent;
	}
	
	return null;
}

sdItem.prototype.addDescription = function(text) {
	var desc = this.getTextFromReference(this.item, 'aria-describedby');
	if (desc) {
		text = text + ' ... description: ' + desc;
	} else if (this.desc) {
		
				text = text + ' ... description: ' + this.desc.nameFromItem();
			
	}
	
	return text;
}

sdItem.prototype.nameFromRole = function() {
	var text = this.item.getAttribute('role');
	if (text && text.indexOf('graphics-') == 0) {
		text = text.substring(9);
	}
	if (this.item.hasAttribute('aria-gtype')) {
		text = text + " " + this.item.getAttribute('aria-gtype');
	}
	
	return text;
}

// Map for objects 
function dmap() {
	this.keys = [];
	this.values = [];
}

// Put something in the map.
dmap.prototype.put = function(key, value) {
	this.keys.push(key);
	this.values.push(value);
}

// Get something from the map.
dmap.prototype.get = function(key) {
	var index = this.keys.indexOf(key);
	return (index == -1) ? null : this.values[index];
}

dmap.prototype.getArray = function(keys) {
	var values = [];
	for(var i = 0; i < keys.length; i++) {
		var value = this.get(keys[i]);
		values.push(value);
	}
	
	return values;
}

// Returns true if the map contains the key.
dmap.prototype.contains = function(key) {
	return this.keys.indexOf(key) != -1;
}




